{- «•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»

    Copyright © 2011, Ingo Wechsung
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

        Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

        Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials provided
        with the distribution. Neither the name of the copyright holder
        nor the names of its contributors may be used to endorse or
        promote products derived from this software without specific
        prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE
    COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
    USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
    IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
    THE POSSIBILITY OF SUCH DAMAGE.

    «•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•»«•» -}

{--
    Handling of constants (literals).
    
    -}

module frege.compiler.gen.Const where

import Data.TreeMap (insertkv)

import frege.compiler.Data as D
import frege.compiler.Utilities as U()

--- java type name of the inner static class that holds the literals
constClass g = "Ĳ"

--- the name of the constant class instance is ĳ
ijLigatur = "ĳ"


{--
     Find the 'JName' of a literal.
     If this literal does not have a name yet, it will be entered in the 'GenSt.consts'
     table.
 -}
findConst (Lit {pos,kind,value})
     | kind == LBool = case value of
         "true" -> stio (JName "Box.Bool" "t")
         _      -> stio (JName "Box.Bool" "f")
     | otherwise  = do
         g <- getST
         case g.gen.consts.lookup (kind,value) of
             Just u
                 | kind `elem` [LInt, LLong, LBig]
                             -> stio (JName ijLigatur (kinda kind ++ value))
                 | otherwise -> stio (JName ijLigatur (kinda kind ++ show u))
             Nothing
                 | kind `elem` [LInt, LLong, LBig] = do
                      changeST Global.{gen <- GenSt.{consts <- insertkv (kind, value) 0}}
                      stio (JName ijLigatur (kinda kind ++ value))
                 | otherwise = do
                      u <- U.uniqid
                      changeST Global.{gen <- GenSt.{consts <- insertkv (kind, value) u}}
                      stio (JName ijLigatur (kinda kind ++ show u))
 
     where
         kinda LBool   = "bool"
         kinda LInt    = "int"
         kinda LRegex  = "rgx"
         kinda LString = "str"
         kinda LDouble = "dbl"
         kinda LFloat  = "flt"
         kinda LChar   = "chr"
         kinda LBig    = "big"
         kinda LLong   = "long"
findConst _ = error "findConst: no Lit"
 


