
{-- 
    'Identity', along with instances for 'Functor',
    'Applicative', 'Bind', 'Apply', 'Monad', 
    'Semigroup', 'Monoid' and 'ListSource'
    -}

package frege.data.wrapper.Identity where

import Data.Monoid

{--
  
    The identity functor and monad.

    This trivial type constructor serves two purposes:

    - It can be used with functions parameterized by a 
      'Functor' or 'Monad'.

    - It can be used as a base monad to which a series of monad
      transformers may be applied to construct a composite monad.
      Most monad transformer modules include the special case of
      applying the transformer to 'Identity'.  For example, @State s@
      is an abbreviation for @StateT s 'Identity'@. 
-}


--- Identity functor and monad.
data Identity a = Identity { run :: a }


instance Monad Identity where
    m >>= k  = k (Identity.run m)
    pure  a = Identity a
    Identity f <*> Identity x = Identity (f x)
    fmap f m = Identity (f (Identity.run m))


instance Semigroup Semigroup a => Identity a where
   Identity x `mappend` Identity y = Identity (x `mappend` y)

instance Monoid Monoid a => Identity a where
   mempty = Identity mempty 
   
instance ListSource Identity where
   toList (Identity x) = [x]   
   